#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# License: BSD
#   https://raw.githubusercontent.com/stonier/groot_rocker_extensions/devel/LICENSE
#

##############################################################################
# Imports
##############################################################################

import os
import shlex

from . import utilities

##############################################################################
# Tests
##############################################################################


class Ssh(utilities.ExtensionTestCase):

    # disabled until persistent + SSH_AUTH_SOCK solution sis found
    def do_not_ssh_auth_sock(self):
        mock_cli_args = {}

        # with SSH_AUTH_SOCK set
        os.environ['SSH_AUTH_SOCK'] = 'foo'
        args = self.extension.get_docker_args(mock_cli_args)
        string_to_check = '-e SSH_AUTH_SOCK --volume ' + shlex.quote('{SSH_AUTH_SOCK}:{SSH_AUTH_SOCK}'.format(**os.environ))
        utilities.assert_details(
            text="ssh_auth_sock: detect and set run args",
            expected=True,
            result=string_to_check in args
        )
        self.assertIn(string_to_check, args)

        # without SSH_AUTH_SOCK set
        del os.environ['SSH_AUTH_SOCK']
        args = self.extension.get_docker_args(mock_cli_args)
        utilities.assert_details(
            text="ssh_auth_sock: detect not present",
            expected=False,
            result="SSH_AUTH_SOCK" in args
        )
        self.assertNotIn('SSH_AUTH_SOCK', args)



#!/usr/bin/env python
#
# License: BSD
#   https://raw.githubusercontent.com/stonier/groot_rocker_extensions/devel/LICENSE
#
##############################################################################
# Documentation
##############################################################################

"""
Transfer a user's git configuration to the docker container. This extension
really only makes sense with the 'user' extension, but won't fail if that
has not been activated.
"""

##############################################################################
# Imports
##############################################################################

import os
import re
import typing

import groot_rocker

##############################################################################
# Extension
##############################################################################


class Git(groot_rocker.extensions.RockerExtension):
    """
    Transfers system and user (if the user extension has been selected) git
    configuration files to the docker container as volumes.
    """
    @classmethod
    def get_name(cls) -> str:
        return re.sub(r'(?<!^)(?=[A-Z])', '_', cls.__name__).lower()  # CamelCase to underscores

    def precondition_environment(self, unused_cli_args: typing.Dict[str, str]):
        pass

    def validate_environment(self, unused_cli_args: typing.Dict[str, str]):
        pass

    def get_preamble(self, unused_cli_args: typing.Dict[str, str]) -> str:
        return ''

    def get_snippet(self, unused_cli_args: typing.Dict[str, str]) -> str:
        return ''

    def get_docker_args(self, cli_args: typing.Dict[str, str]) -> str:
        args = ''
        system_gitconfig = '/etc/gitconfig'
        user_gitconfig = os.path.expanduser('~/.gitconfig')
        user_gitconfig_target = '/root/.gitconfig'
        if 'user' in cli_args and cli_args['user']:
            user_gitconfig_target = user_gitconfig
        if os.path.exists(system_gitconfig):
            args += ' -v {system_gitconfig}:{system_gitconfig}:ro'.format(**locals())
        if os.path.exists(user_gitconfig):
            args += ' -v {user_gitconfig}:{user_gitconfig_target}:ro'.format(**locals())
        return args

    @staticmethod
    def register_arguments(parser, defaults={}):
        parser.add_argument(
            '--git',
            action='store_true',
            default=defaults.get('git', None),
            help="use host settings (/etc/gitconfig and ~/.gitconfig)"
        )



#!/usr/bin/env python
#
# License: BSD
#   https://raw.githubusercontent.com/stonier/groot_rocker_extensions/devel/LICENSE
#
##############################################################################
# Documentation
##############################################################################
"""
Export a coloured, named prompt to ~/.bash_profile on container creation.
"""

##############################################################################
# Imports
##############################################################################

import em
import pkgutil
import os
import pwd
import re
import typing

import groot_rocker

##############################################################################
# Extension
##############################################################################


class NamedPrompt(groot_rocker.extensions.RockerExtension):

    @classmethod
    def get_name(cls) -> str:
        return re.sub(r'(?<!^)(?=[A-Z])', '_', cls.__name__).lower()  # CamelCase to underscores

    def precondition_environment(self, unused_cli_args: typing.Dict[str, str]):
        pass

    def validate_environment(self, unused_cli_args: typing.Dict[str, str]):
        pass

    @staticmethod
    def desired_extensions():
        return {"user"}

    def get_preamble(self, unused_cli_args: typing.Dict[str, str]) -> str:
        return ''

    def get_snippet(self, cli_args: typing.Dict[str, str]) -> str:
        snippet = pkgutil.get_data(
            'groot_rocker_extensions',
            'templates/named_prompt.Dockerfile.em'
        ).decode('utf-8')
        substitutions = {}
        if 'user' in cli_args and cli_args['user']:
            userinfo = pwd.getpwuid(os.getuid())
            substitutions['user_name'] = getattr(userinfo, 'pw_' + 'name')
            substitutions['home_dir'] = f"/home/{substitutions['user_name']}"
        else:
            substitutions['user_name'] = "root"
            substitutions['home_dir'] = "/root"
        if 'container_name' in cli_args and cli_args['container_name']:
            substitutions['container_name'] = cli_args['container_name']
        else:
            substitutions['container_name'] = r'\h'
        return em.expand(snippet, substitutions)

    def get_docker_args(self, unused_cli_args: typing.Dict[str, str]) -> str:
        return ""

    @staticmethod
    def register_arguments(parser, defaults={}):
        # TODO: what to do with the defaults arg?
        parser.add_argument(
            '--named-prompt',
            action='store_true',
            help='export a named prompt via PS1 to ~/.bash_profile'
        )



#!/usr/bin/env python
#
# License: BSD
#   https://raw.githubusercontent.com/stonier/groot_rocker_extensions/devel/LICENSE
#
##############################################################################
# Documentation
##############################################################################

"""
Console scripts that provide user conveniences.
"""

##############################################################################
# Imports
##############################################################################

import argparse
import sys

import groot_rocker
import groot_rocker.console as console

##############################################################################
# Sandbox
##############################################################################


def main_workspace():
    defaults = {
        "command": "/bin/bash --login -i",
        "development_environment": True,
        "extension_blacklist": [],
        "git": True,
        "mode": "interactive",
        "named_prompt": True,
        "network": "host",
        "persistent": True,
        "user": True,
        "ssh": True,
    }
    parser = argparse.ArgumentParser(
        description=(
            "Build and enter a sandboxed docker environment for development with the\n"
            "following extensions enabled:\n"
            " - development_environment\n"
            " - git\n"
            " - named_prompt\n"
            " - network (=host)\n"
            " - persistent\n"
            " - user\n"
            " - ssh\n"
            "It also sets the command for an interactive bash login."
        ),
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    build_options = parser.add_argument_group(title="Build Options", description="Options for image/container creation.")
    build_options.add_argument(
        '--name', type=str,
        default="development",
        help="human readable name for both image (sandbox:<name>) & container"
    )

    extensions_group = parser.add_argument_group(title="Extensions")
    extensions = ["bind", "colcon", "nvidia", "pulse_audio", "work_directory"]
    available_extensions = groot_rocker.core.list_plugins()  # typing.Dict[str, Extension]
    print(f"Exts: {available_extensions}")
    for extension in extensions:
        available_extensions[extension].register_arguments(extensions_group, defaults)

    parser.add_argument(
        'image', nargs='?',
        default=groot_rocker.cli.set_default("image", defaults),
        help="base image to build from (required)"
    )
    options = vars(parser.parse_args())  # work with a dict object from here, not argparse.Namespace
    options['container_name'] = options['name']
    options['image_name'] = f"devel:{options['name']}"
    del options['name']
    options.update(defaults)
    if options['image'] is None:
        console.error("No image was specified (hint: did you use --bind and not provide a '--' to close it's arguments?)")
        sys.exit(1)
    # convenience setting of work dir if there is only one mount point
    if options['work_directory'] is None and (len(options['bind']) == 1):
        bind_argument = options['bind'][0]
        if ":" in bind_argument:
            unused_source, options['work_directory'] = bind_argument.split(":")
        else:
            options['work_directory'] = bind_argument
    groot_rocker.cli.build_and_run(options)



#!/usr/bin/env python
#
# License: BSD
#   https://raw.githubusercontent.com/stonier/groot_rocker_extensions/devel/LICENSE
#
##############################################################################
# Documentation
##############################################################################

"""
Forward PulseAudio configuration into the container. Only tested in Ubuntu 18.
"""

##############################################################################
# Imports
##############################################################################

import em
import grp
import os
import pathlib
import pkgutil
import pwd
import re
import typing

import groot_rocker

##############################################################################
# Extension
##############################################################################


class PulseAudio(groot_rocker.extensions.RockerExtension):
    """
    Forward PulseAudio configuration into the container.
    This will not work without --user.

    This extension requires a logged in desktop user with
    the defualt pulse audio configuration. One problem with this is that
    the exact same user permissions are required within and without, which
    makes it unsuitable for headless systems.

    This can be worked around, but it likely requires some level of
    host configuration. Refer to

      https://github.com/stonier/groot_rocker_extensions/issues/31

    Regardless, this least-effort implementation is still desirable
    and useful for logged in desktop users.

    Test with the 'speaker-test' utility that gets installed with this
    extension.

    .. code=block:: bash

       speaker-test --channels=2 --nloops=1 --test=wav --device=default

    """
    @classmethod
    def get_name(cls) -> str:
        return re.sub(r'(?<!^)(?=[A-Z])', '_', cls.__name__).lower()  # CamelCase to underscores

    def precondition_environment(self, unused_cli_args: typing.Dict[str, str]):
        pass

    def validate_environment(self, unused_cli_args: typing.Dict[str, str]):
        pulse_dir = f"/run/user/{os.getuid()}/pulse"
        if not os.path.exists(pulse_dir):
            raise groot_rocker.core.ValidateError(
                f"user's pulseaudio not found [{pulse_dir}]"
            )

    @staticmethod
    def required_extensions():
        return {"user"}

    def get_preamble(self, unused_cli_args: typing.Dict[str, str]) -> str:
        return ''

    def get_snippet(self, cli_args: typing.Dict[str, str]) -> str:
        container_pulse_socket = os.path.join(os.getenv('XDG_RUNTIME_DIR'), "pulse", "native")
        snippet = pkgutil.get_data(
            'groot_rocker_extensions',
            f"templates/{PulseAudio.get_name()}_snippet.Dockerfile.em").decode('utf-8')
        substitutions = {
            "audio_dependencies": (
                "alsa-utils libasound2-plugins"
            ),
            "pulse_socket": container_pulse_socket,
            # 'user' is a required extension
            "user_active": True,
            "user_id": os.getuid(),
            "user_name": getattr(pwd.getpwuid(os.getuid()), 'pw_' + 'name')
        }
        return em.expand(snippet, substitutions)

    def get_docker_args(self, cli_args: typing.Dict[str, str]) -> str:
        # XDG_RUNTIME_DIR
        #   defn: location for a user's runtime artifacts as spec'd by freedesktop
        #   note: this requires a host system with a logged in user as it is only
        #         configured upon login
        host_user_runtime_dir = pathlib.Path(os.getenv('XDG_RUNTIME_DIR'))
        substitutions = {
            "audio_group_id": f"{grp.getgrnam('audio').gr_gid}",
            "host_user_pulse_dir": os.path.join(host_user_runtime_dir, "pulse")
        }
        args = (
            " --device /dev/snd "
            " -v /usr/share/alsa:/usr/share/alsa "
            " -v %(host_user_pulse_dir)s:%(host_user_pulse_dir)s "
            " -e PULSE_SERVER=unix:%(host_user_pulse_dir)s/native "
            " --group-add %(audio_group_id)s "
        )
        return args % substitutions

    @staticmethod
    def register_arguments(parser, defaults={}):
        parser.add_argument(
            '--pulse-audio',
            action='store_true',
            default=defaults.get(PulseAudio.get_name(), None),
            help="ride along on the host user's pulseaudio"
        )



#!/usr/bin/env python
#
# License: BSD
#   https://raw.githubusercontent.com/stonier/groot_rocker_extensions/devel/LICENSE
#
##############################################################################
# Documentation
##############################################################################

"""
Switch to the specified working directory on entry into the container.
"""

##############################################################################
# Imports
##############################################################################

import em
import os
import pkgutil
import re
import typing

import groot_rocker

##############################################################################
# Extension
##############################################################################


class WorkDirectory(groot_rocker.extensions.RockerExtension):
    """
    Be robust to '~', but otherwise just directly pass in the
    specified working directory to the dockerfile WORKDIR variable.

    .. todo:: check that the working directory actually exists in the image
    """

    @classmethod
    def get_name(cls) -> str:
        return re.sub(r'(?<!^)(?=[A-Z])', '_', cls.__name__).lower()  # CamelCase to underscores

    def precondition_environment(self, unused_cli_args: typing.Dict[str, str]):
        pass

    def validate_environment(self, unused_cli_args: typing.Dict[str, str]):
        pass

    def get_preamble(self, unused_cli_args: typing.Dict[str, str]):
        return ''

    def get_snippet(self, cli_args: typing.Dict[str, str]):
        work_directory = cli_args[WorkDirectory.get_name()]
        if work_directory == "~":  # expand
            work_directory = os.path.expanduser("~")
        snippet = pkgutil.get_data(
            "groot_rocker_extensions",
            f"templates/{WorkDirectory.get_name()}.Dockerfile.em"
        ).decode('utf-8')
        substitutions = {WorkDirectory.get_name(): work_directory}
        return em.expand(snippet, substitutions)

    def get_docker_args(self, unused_cli_args: typing.Dict[str, str]):
        return ''

    @staticmethod
    def register_arguments(parser, defaults={}):
        parser.add_argument(
            '--work-directory',
            type=str,
            default=defaults.get(WorkDirectory.get_name(), None),
            help='set the default work directory on container entry'
        )



#!/usr/bin/env python
#
# License: BSD
#   https://raw.githubusercontent.com/stonier/groot_rocker_extensions/devel/LICENSE
#
##############################################################################
# Documentation
##############################################################################

"""
Mount host paths in the docker container via docker's bind mechanism.
"""

##############################################################################
# Imports
##############################################################################

import os
import re
import typing

import groot_rocker

##############################################################################
# Extension
##############################################################################


class Bind(groot_rocker.extensions.RockerExtension):
    """
    Mounts host paths to target paths in the container via the docker bind
    mechanism.

    .. note:

       Volumes are preferred in a docker-verse, but these cannot operate on
       existing host paths. Bind however, comes with the entanglement of
       permissions.

    .. tip:

       In most cases, you would use this extension with the
      'user' extension to ensure internal and external permissions align.
    """

    @classmethod
    def get_name(cls) -> str:
        return re.sub(r'(?<!^)(?=[A-Z])', '_', cls.__name__).lower()  # CamelCase to underscores

    def precondition_environment(self, unused_cli_args: typing.Dict[str, str]):
        pass

    def validate_environment(self, unused_cli_args: typing.Dict[str, str]):
        pass

    def get_preamble(self, unused_cli_args: typing.Dict[str, str]) -> str:
        return ''

    def get_snippet(self, unused_cli_args: typing.Dict[str, str]) -> str:
        return ''

    def get_docker_args(self, cli_args: typing.Dict[str, str]) -> str:
        args = []
        for mount in cli_args['bind']:
            for d in mount.split(" "):
                if ":" in d:
                    src, dst = d.split(":")
                else:
                    src = d
                    dst = d
                src = os.path.abspath(src)
                args.append(f"--mount type=bind,source={src},target={dst}")
        return ' ' + ' '.join(args)

    @staticmethod
    def register_arguments(parser, defaults={}):
        # Be a little robust to user input defaults. Very often there is just
        # one argument, this allows people to pass it in as a string, not a list
        default = defaults.get(Bind.get_name(), [])
        if isinstance(default, str):
            default = [default]
        parser.add_argument(
            '--bind',
            type=str,
            nargs='+',
            metavar="SOURCE:TARGET",
            default=default,
            help="bind host folders at arbitrary mount points"
        )



#!/usr/bin/env python
#
# License: BSD
#   https://raw.githubusercontent.com/stonier/groot_rocker_extensions/devel/LICENSE
#
##############################################################################
# Documentation
##############################################################################

"""
Install colcon build tools.
"""

##############################################################################
# Imports
##############################################################################

import em
import os
import pkgutil
import pwd
import re
import typing

import groot_rocker

##############################################################################
# Extension
##############################################################################


class Colcon(groot_rocker.extensions.RockerExtension):
    """
    Installs tooling to support development with colcon.
    """
    @classmethod
    def get_name(cls) -> str:
        return re.sub(r'(?<!^)(?=[A-Z])', '_', cls.__name__).lower()  # CamelCase to underscores

    @staticmethod
    def desired_extensions():
        return {"user"}

    def get_files(self, unused_cli_args):
        templates_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), "templates")
        colcon_setup_path = os.path.join(templates_path, f"{Colcon.get_name()}_setup.bash")
        file = open(colcon_setup_path, "r")
        contents = file.read()
        file.close()
        return {"colcon_setup.bash": contents}

    def get_snippet(self, cli_args: typing.Dict[str, str]) -> str:
        snippet = pkgutil.get_data(
            "groot_rocker_extensions",
            f"templates/{Colcon.get_name()}.Dockerfile.em"
        ).decode('utf-8')

        substitutions = {}
        if 'user' in cli_args and cli_args['user']:
            substitutions['user_active'] = True
            substitutions["user_name"] = getattr(pwd.getpwuid(os.getuid()), 'pw_' + 'name')
        else:
            substitutions['user_active'] = False
        return em.expand(snippet, substitutions)

    @staticmethod
    def register_arguments(parser, defaults={}):
        parser.add_argument(
            '--colcon',
            action='store_true',
            default=defaults.get(Colcon.get_name(), None),
            help="install the colcon build tools"
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# License: BSD
#   https://raw.githubusercontent.com/stonier/groot_rocker_extensions/devel/LICENSE
#

##############################################################################
# Imports
##############################################################################

import os

from . import utilities

##############################################################################
# Tests
##############################################################################


class PulseAudio(utilities.ExtensionTestCase):

    def test_pulse_server_env_variable(self):
        mock_cli_args = {}

        os.environ['XDG_RUNTIME_DIR'] = '/run/user/1000'

        args = self.extension.get_docker_args(mock_cli_args)
        string_to_check = '-e PULSE_SERVER=unix:'
        utilities.assert_details(
            text="Set PULSE_SERVER Env Variable",
            expected=True,
            result=string_to_check in args
        )
        self.assertIn(string_to_check, args)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# License: BSD
#   https://raw.githubusercontent.com/stonier/groot_rocker_extensions/devel/LICENSE
#

##############################################################################
# Imports
##############################################################################

import os

from . import utilities

##############################################################################
# Tests
##############################################################################


class Git(utilities.ExtensionTestCase):

    def test_detection(self):
        """
        This is not an ideal test - it depends on whether or not the
        system that runs the test has git configuration or not.
        Ideally, that ought to be mocked in some way. At the least though
        it provides a useful development test with the developer's
        system.
        """
        mock_cli_args = {}
        self.assertEqual(self.extension.get_snippet(mock_cli_args), '')
        self.assertEqual(self.extension.get_preamble(mock_cli_args), '')
        args = self.extension.get_docker_args(mock_cli_args)

        system_gitconfig = '/etc/gitconfig'
        user_gitconfig = os.path.expanduser('~/.gitconfig')
        user_gitconfig_target = '/root/.gitconfig'
        utilities.assert_details(
            text=f"Mount {system_gitconfig}",
            expected=os.path.exists(system_gitconfig),
            result=(f"-v {system_gitconfig}:{system_gitconfig}" in args)
        )
        if os.path.exists(system_gitconfig):
            self.assertIn('-v %s:%s' % (system_gitconfig, system_gitconfig), args)
        utilities.assert_details(
            text=f"Mount {user_gitconfig_target}",
            expected=os.path.exists(user_gitconfig),
            result=(f"-v {user_gitconfig}:{user_gitconfig_target}" in args)
        )
        if os.path.exists(user_gitconfig):
            self.assertIn('-v %s:%s' % (user_gitconfig, user_gitconfig_target), args)

        # Test with user "enabled"
        mock_cli_args = {'user': True}
        user_args = self.extension.get_docker_args(mock_cli_args)
        utilities.assert_details(
            text=f"Mount {user_gitconfig}",
            expected=os.path.exists(system_gitconfig),
            result=(f"-v {user_gitconfig}:{user_gitconfig}" in args)
        )
        if os.path.exists(user_gitconfig):
            self.assertIn('-v %s:%s' % (user_gitconfig, user_gitconfig), user_args)



#!/usr/bin/env python
#
# License: BSD
#   https://raw.githubusercontent.com/stonier/groot_rocker_extensions/devel/LICENSE
#
##############################################################################
# Documentation
##############################################################################

"""
Switch to the specified working directory on entry into the container.
"""

##############################################################################
# Imports
##############################################################################

import em
import os
import pkgutil
import re
import shlex
import typing

import groot_rocker

##############################################################################
# Extension
##############################################################################


class Ssh(groot_rocker.extensions.RockerExtension):
    """
    Does a few ssh related activities:

     * Image: ensures ssh server and client debians are installed.
     * Container: forwards the SSH_AUTH_SOCK agent properties into the container
     * Container: if the user extension has been applied and ~/.ssh exists, mount it
    """

    @classmethod
    def get_name(cls) -> str:
        return re.sub(r'(?<!^)(?=[A-Z])', '_', cls.__name__).lower()  # CamelCase to underscores

    def precondition_environment(self, unused_cli_args: typing.Dict[str, str]):
        pass

    def validate_environment(self, unused_cli_args: typing.Dict[str, str]):
        pass

    def get_preamble(self, unused_cli_args: typing.Dict[str, str]):
        return ''

    def get_snippet(self, unused_cli_args: typing.Dict[str, str]):
        snippet = pkgutil.get_data(
            "groot_rocker_extensions",
            f"templates/{Ssh.get_name()}.Dockerfile.em"
        ).decode('utf-8')
        return em.expand(snippet, {})

    def get_docker_args(self, cli_args: typing.Dict[str, str]):
        """
        It would be useful to enable SSH_AUTH_SOCK for passwordless logins.
        This, however, fails with persistent containers since this hardcodes SSH_AUTH_SOCK
        at the time of calling docker run to create the container. If the host reboots,
        the location is no longer valid and a call to 'container start <name> -i' will
        fail as reported in #27.
        """
        args = ''
        # if 'SSH_AUTH_SOCK' in os.environ:
        #     args += " -e SSH_AUTH_SOCK"
        #     args += " --volume " + shlex.quote("{SSH_AUTH_SOCK}:{SSH_AUTH_SOCK}".format(**os.environ))
        return args

    @staticmethod
    def register_arguments(parser, defaults={}):
        parser.add_argument(
            '--ssh',
            action="store_true",
            default=defaults.get(Ssh.get_name(), None),
            help='configures the container with ssh capabilities'
        )
